Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Lwt_eio.PromiseSourceawait_lwt p allows an Eio fiber to wait for a Lwt promise p to be resolved, much like Eio.Promise.await does for Eio promises.
This can only be used while the event loop created by with_event_loop is still running.
Note: in the usual case of needing to run some Lwt code to generate the promise first, use run_lwt instead. That handles cancellation and also tracks the current context so that debug mode works.
await_eio p allows a Lwt thread to wait for an Eio promise p to be resolved. This can only be used while the event loop created by with_event_loop is still running.
await_eio_result is like await_eio, but allows failing the lwt thread too.